// Scintilla source code edit control
// ScintillaFOX.cxx - FOX toolkit specific subclass of ScintillaBase
// Copyright 1998-2001 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be distributed.

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>

#include <fox/fx.h>
#include <fox/fxkeys.h>

#include "Platform.h"

#include "Scintilla.h"
#include "ScintillaWidget.h"
#ifdef SCI_LEXER
#include "SciLexer.h"
#include "PropSet.h"
#include "Accessor.h"
#include "KeyWords.h"
#endif
#include "ContractionState.h"
#include "SVector.h"
#include "CellBuffer.h"
#include "CallTip.h"
#include "KeyMap.h"
#include "Indicator.h"
#include "LineMarker.h"
#include "Style.h"
#include "AutoComplete.h"
#include "ViewStyle.h"
#include "Document.h"
#include "Editor.h"
#include "SString.h"
#include "ScintillaBase.h"

#include <fstream.h>	// <FIXME/>

#ifdef _MSC_VER
// Constant conditional expressions are because of GTK+ headers
#pragma warning(disable: 4127)
#endif

static FXWindow *PWidget(Window & w) {
	return reinterpret_cast<FXWindow *>(w.GetID());
}

static FXMenuPane *PMenu(WindowID id) {
	return reinterpret_cast<FXMenuPane *>(id);
}

#define ID(id) (FXScrollArea::ID_LAST+id)

// ====================================================================
// SCPopupMenuCommand
// --------------------------------------------------------------------
// <FIXME> Maybe not needed anymore coz FOX should handle right button
// release correctly since 0.99.176 </FIXME>
// ====================================================================
class SCPopupMenuCommand : public FXMenuCommand
{
	FXDECLARE(SCPopupMenuCommand)
protected:
	SCPopupMenuCommand() {}
public:
	long onRightBtnRelease(FXObject *, FXSelector, void *);
public:
  SCPopupMenuCommand(FXComposite* p,const FXString& text,FXIcon* ic=NULL,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0);
};

// Map
FXDEFMAP(SCPopupMenuCommand) SCPopupMenuCommandMap[]={
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,SCPopupMenuCommand::onRightBtnRelease),
  };

// Object implementation
FXIMPLEMENT(SCPopupMenuCommand,FXMenuCommand,SCPopupMenuCommandMap,ARRAYNUMBER(SCPopupMenuCommandMap))

// Command menu item
SCPopupMenuCommand::SCPopupMenuCommand(FXComposite* p,const FXString& text,FXIcon* ic,FXObject* tgt,FXSelector sel,FXuint opts):
FXMenuCommand(p, text, ic, tgt, sel, opts)
{}

// Released right button
long SCPopupMenuCommand::onRightBtnRelease(FXObject*,FXSelector,void* ptr){
  FXbool active=isActive();
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr)) return 1;
    getParent()->handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),NULL);
    if(active){
      if(target) target->handle(this,MKUINT(message,SEL_COMMAND),(void*)1);
    }
    return 1;
  }
  return 0;
}

// ====================================================================
// ScintillaFOX
// ====================================================================
class ScintillaFOX : public FXScrollArea, public ScintillaBase {
  FXDECLARE(ScintillaFOX)
protected:
	bool tryDrag;
	bool dragWasDropped;
	Point ptMouseLastBeforeDND;

	// Because clipboard access is asynchronous, copyText is created by Copy
	SelectionText copyText;
	SelectionText primary;
	
protected:
	ScintillaFOX() {}
public:
	enum {
		idLast = ID(100),
		ID_TICK,
	};
public:
	long onScintillaCommand(FXObject *, FXSelector, void*);
	long onPaint(FXObject *, FXSelector, void *);
	long onConfigure(FXObject *, FXSelector, void *);
	long onTimeoutTicking(FXObject *, FXSelector, void *);
	long onFocusIn(FXObject *, FXSelector, void *);
	long onFocusOut(FXObject *, FXSelector, void *);
	long onMotion(FXObject *, FXSelector, void *);
	long onLeftBtnPress(FXObject *, FXSelector, void *);
	long onLeftBtnRelease(FXObject *, FXSelector, void *);
	long onRightBtnPress(FXObject *, FXSelector, void *);
	long onMiddleBtnPress(FXObject *, FXSelector, void *);
	long onBeginDrag(FXObject *, FXSelector, void *);
	long onDragged(FXObject *, FXSelector, void *);
	long onEndDrag(FXObject *, FXSelector, void *);
	long onDNDEnter(FXObject *, FXSelector, void *);
	long onDNDLeave(FXObject *, FXSelector, void *);
	long onDNDMotion(FXObject *, FXSelector, void *);
	long onDNDDrop(FXObject *, FXSelector, void *);
	long onDNDRequest(FXObject *, FXSelector, void *);
	long onSelectionLost(FXObject *, FXSelector, void *);
	long onSelectionRequest(FXObject *, FXSelector, void *);
	long onClipboardLost(FXObject *, FXSelector, void *);
	long onClipboardRequest(FXObject *, FXSelector, void *);
	long onKeyPress(FXObject *, FXSelector, void *);
public:
	ScintillaFOX(FXComposite * p, FXObject * target = NULL, FXSelector sel = 0);
	virtual ~ScintillaFOX();
	virtual sptr_t WndProc(unsigned int iMessage, uptr_t wParam, sptr_t lParam);
private:
	// Virtual methods from ScintillaBase
	virtual void Initialise();
	virtual void Finalise();
	virtual void SetVerticalScrollPos();
	virtual void SetHorizontalScrollPos();
	virtual void Copy();
	virtual void Paste();
	virtual void ClaimSelection();
	virtual void NotifyChange();
	virtual void NotifyParent(SCNotification scn);
	virtual void SetTicking(bool on);
	virtual void SetMouseCapture(bool on);
	virtual bool HaveMouseCapture();
	virtual sptr_t DefWndProc(unsigned int iMessage, uptr_t wParam, sptr_t lParam);
	virtual void CreateCallTipWindow(PRectangle rc);
	virtual void AddToPopUp(const char * label, int cmd = 0, bool enabled = true);
	virtual void StartDrag();
	//
	static sptr_t DirectFunction(ScintillaFOX *sciThis, 
		unsigned int iMessage, uptr_t wParam, sptr_t lParam);
	// Paint
	void SyncPaint(PRectangle rc);
	void FullPaint();
	void UnclaimSelection();
	void ReceivedSelection(FXDNDOrigin origin);
	void NotifyKey(int key, int modifiers);
	void NotifyURIDropped(const char *list);
	
	// From FXScrollArea
	virtual void create();
  virtual FXbool canFocus() const;
	virtual FXint getContentWidth();
	virtual FXint getContentHeight();
  virtual void moveContents(FXint x,FXint y);

	// From ScintillaBase
	virtual bool ModifyScrollBars(int nMax, int nPage);
	virtual void ReconfigureScrollBars();
	virtual PRectangle GetClientRectangle();
	virtual int KeyDefault(int key, int modifiers);
};

FXDEFMAP(ScintillaFOX) ScintillaFOXMap[]={
  FXMAPFUNCS(SEL_COMMAND, FXScrollArea::ID_LAST, ScintillaFOX::idLast, ScintillaFOX::onScintillaCommand),
  FXMAPFUNC(SEL_PAINT, 0, ScintillaFOX::onPaint),
  FXMAPFUNC(SEL_CONFIGURE,0,ScintillaFOX::onConfigure),
  FXMAPFUNC(SEL_TIMEOUT,ScintillaFOX::ID_TICK,ScintillaFOX::onTimeoutTicking),
  FXMAPFUNC(SEL_FOCUSIN,0,ScintillaFOX::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,ScintillaFOX::onFocusOut),
  FXMAPFUNC(SEL_MOTION,0,ScintillaFOX::onMotion),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,ScintillaFOX::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,ScintillaFOX::onLeftBtnRelease),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,ScintillaFOX::onRightBtnPress),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,ScintillaFOX::onMiddleBtnPress),
  FXMAPFUNC(SEL_BEGINDRAG,0,ScintillaFOX::onBeginDrag),
  FXMAPFUNC(SEL_DRAGGED,0,ScintillaFOX::onDragged),
  FXMAPFUNC(SEL_ENDDRAG,0,ScintillaFOX::onEndDrag),
  FXMAPFUNC(SEL_DND_ENTER,0,ScintillaFOX::onDNDEnter),
  FXMAPFUNC(SEL_DND_LEAVE,0,ScintillaFOX::onDNDLeave),
  FXMAPFUNC(SEL_DND_DROP,0,ScintillaFOX::onDNDDrop),
  FXMAPFUNC(SEL_DND_MOTION,0,ScintillaFOX::onDNDMotion),
  FXMAPFUNC(SEL_DND_REQUEST,0,ScintillaFOX::onDNDRequest),
  FXMAPFUNC(SEL_SELECTION_LOST,0,ScintillaFOX::onSelectionLost),
//  FXMAPFUNC(SEL_SELECTION_GAINED,0,ScintillaFOX::onSelectionGained),
  FXMAPFUNC(SEL_SELECTION_REQUEST,0,ScintillaFOX::onSelectionRequest),
  FXMAPFUNC(SEL_CLIPBOARD_LOST,0,ScintillaFOX::onClipboardLost),
//  FXMAPFUNC(SEL_CLIPBOARD_GAINED,0,ScintillaFOX::onClipboardGained),
  FXMAPFUNC(SEL_CLIPBOARD_REQUEST,0,ScintillaFOX::onClipboardRequest),
  FXMAPFUNC(SEL_KEYPRESS,0,ScintillaFOX::onKeyPress),
};

FXIMPLEMENT(ScintillaFOX,FXScrollArea,ScintillaFOXMap,ARRAYNUMBER(ScintillaFOXMap))

ScintillaFOX::ScintillaFOX(FXComposite * p, FXObject * tgt, FXSelector sel) :
	FXScrollArea(p, LAYOUT_FILL_X|LAYOUT_FILL_Y),
	tryDrag(false),
	dragWasDropped(false)
{
  flags|=FLAG_ENABLED;
	wMain = this;
	target = tgt;
	message = sel;
	Initialise();
}

ScintillaFOX::~ScintillaFOX()
{
	Finalise();
}

void ScintillaFOX::create()
{
  FXScrollArea::create();
  if(!textType){textType=getApp()->registerDragType(textTypeName);}
  if(!urilistType){urilistType=getApp()->registerDragType(urilistTypeName);}
	dropEnable();
}

FXbool ScintillaFOX::canFocus() const
{
	return true;
}

sptr_t ScintillaFOX::WndProc(unsigned int iMessage, uptr_t wParam, sptr_t lParam) {
	switch (iMessage) {

	case SCI_GRABFOCUS:
		setFocus();
		break;

	case SCI_GETDIRECTFUNCTION:
		return reinterpret_cast<sptr_t>(DirectFunction);
	
	case SCI_GETDIRECTPOINTER:
		return reinterpret_cast<sptr_t>(this);

	default:
		return ScintillaBase::WndProc(iMessage,wParam,lParam);
	}
	return 0l;
}

long ScintillaFOX::onScintillaCommand(FXObject * obj, FXSelector sel, void * ptr)
{
	Command(SELID(sel)-ID(0));
	return 1;
}

long ScintillaFOX::onPaint(FXObject * obj, FXSelector sel, void * ptr)
{
	FXScrollArea::onPaint(obj, sel, ptr);

	FXEvent   *ev=(FXEvent*)ptr;
	PRectangle rcPaint(ev->rect.x, ev->rect.y, ev->rect.x + ev->rect.w - 1, ev->rect.y + ev->rect.h - 1);
	SyncPaint(rcPaint);
	return 1;
}

long ScintillaFOX::onTimeoutTicking(FXObject *, FXSelector, void *)
{
	timer.tickerID = FXApp::instance()->addTimeout(timer.tickSize, this, ID_TICK);
	Tick();
	return 1;
}

long ScintillaFOX::onFocusIn(FXObject * sender, FXSelector sel, void * ptr)
{
	FXScrollArea::onFocusIn(sender, sel, ptr);
	SetFocusState(true);
	return 1;
}

long ScintillaFOX::onFocusOut(FXObject * sender, FXSelector sel, void * ptr)
{
	FXScrollArea::onFocusOut(sender, sel, ptr);
	SetFocusState(false);
	return 1;
}

long ScintillaFOX::onMotion(FXObject * sender, FXSelector sel, void * ptr)
{
	FXEvent * ev = static_cast<FXEvent *>(ptr);
	if (tryDrag) {
		tryDrag = false;
		handle(this, MKUINT(0, SEL_BEGINDRAG), 0);
	}
	if (isDragging()) {
		return handle(this, MKUINT(0, SEL_DRAGGED), ptr);
	}
	else {
		Point pt(ev->win_x, ev->win_y);
		ButtonMove(pt);
	}
	return 1;
}

long ScintillaFOX::onLeftBtnPress(FXObject * sender, FXSelector sel, void * ptr)
{
//	if (FXScrollArea::onLeftBtnPress(sender, sel, ptr))
//		return 1;
  handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
	setFocus();
	FXEvent * ev = static_cast<FXEvent *>(ptr);
	Point pt;
	pt.x = ev->win_x;
	pt.y = ev->win_y;
	setFocus();
		//sciThis->ButtonDown(pt, event->time,
		//	event->state & GDK_SHIFT_MASK,
		//	event->state & GDK_CONTROL_MASK,
		//	event->state & GDK_MOD1_MASK);
		// Instead of sending literal modifiers use control instead of alt
		// This is because all the window managers seem to grab alt + click for moving
	ButtonDown(pt, ev->time,
			ev->state & SHIFTMASK,
			ev->state & CONTROLMASK,
			ev->state & CONTROLMASK);
	return 1;
}

long ScintillaFOX::onLeftBtnRelease(FXObject *, FXSelector, void * ptr)
{
	if (isDragging())
		return handle(this, MKUINT(0, SEL_ENDDRAG), ptr);
	FXEvent * ev = static_cast<FXEvent *>(ptr);
	Point pt(ev->win_x, ev->win_y);
	if (!HaveMouseCapture()) {
		if (tryDrag) {
			tryDrag = false;
			SetEmptySelection(PositionFromLocation(pt));
			SetDragPosition(invalidPosition);
		}
		return 1;
	}
	ButtonUp(pt, ev->time, ev->state & CONTROLMASK);
	return 1;
}

long ScintillaFOX::onRightBtnPress(FXObject * sender, FXSelector sel, void * ptr)
{
//	if (FXScrollArea::onRightBtnPress(sender, sel, ptr))
//		return 1;
	if (displayPopupMenu) {
		FXEvent * ev = static_cast<FXEvent *>(ptr);
		ContextMenu(Point(ev->root_x, ev->root_y));
		return 1;
	}
	return 0;
}

long ScintillaFOX::onMiddleBtnPress(FXObject * sender, FXSelector sel, void * ptr)
{
//	if (FXScrollArea::onMiddleBtnPress(sender, sel, ptr))
//		return 1;

	Point pt;
	pt.x = ((FXEvent *)ptr)->win_x;
	pt.y = ((FXEvent *)ptr)->win_y;
	currentPos = PositionFromLocation(pt);
	ReceivedSelection(FROM_SELECTION);
  return 1;
}

// ********************************************************************
// Drag'n drop
// ********************************************************************

// Start a drag operation
long ScintillaFOX::onBeginDrag(FXObject* sender,FXSelector sel,void* ptr){
  SetMouseCapture(true);
  if(FXScrollArea::onBeginDrag(sender,sel,ptr)) return 1;
  beginDrag(&textType,1);
  setDragCursor(getApp()->getDefaultCursor(DEF_DNDSTOP_CURSOR));
  return 1;
  }


// End drag operation
long ScintillaFOX::onEndDrag(FXObject* sender,FXSelector sel,void* ptr){
  SetMouseCapture(false);
  if(FXScrollArea::onEndDrag(sender,sel,ptr)) return 1;
  endDrag((didAccept()!=DRAG_REJECT));
  setDragCursor(getApp()->getDefaultCursor(DEF_TEXT_CURSOR));

	SetDragPosition(invalidPosition);
	return 1;
  }


// Dragged stuff around
long ScintillaFOX::onDragged(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  FXDragAction action;
  if(FXScrollArea::onDragged(sender,sel,ptr)) return 1;
  action=DRAG_COPY;
	if (!pdoc->IsReadOnly()) {
    if(isDropTarget()) action=DRAG_MOVE;
    if(event->state&CONTROLMASK) action=DRAG_COPY;
    if(event->state&SHIFTMASK) action=DRAG_MOVE;
    }
  handleDrag(event->root_x,event->root_y,action);
  if(didAccept()!=DRAG_REJECT){
    if(action==DRAG_MOVE)
      setDragCursor(getApp()->getDefaultCursor(DEF_DNDMOVE_CURSOR));
    else
      setDragCursor(getApp()->getDefaultCursor(DEF_DNDCOPY_CURSOR));
    }
  else{
    setDragCursor(getApp()->getDefaultCursor(DEF_DNDSTOP_CURSOR));
    }
  return 1;
  }


// Handle drag-and-drop enter
long ScintillaFOX::onDNDEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onDNDEnter(sender,sel,ptr);
  return 1;
  }


// Handle drag-and-drop leave
long ScintillaFOX::onDNDLeave(FXObject* sender,FXSelector sel,void* ptr){
  stopAutoScroll();
  FXScrollArea::onDNDLeave(sender,sel,ptr);
  return 1;
  }

// Handle drag-and-drop motion
long ScintillaFOX::onDNDMotion(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXDragAction action;
  FXint pos;

  // Scroll into view
  if(startAutoScroll(ev->win_x,ev->win_y,TRUE)) return 1;

  // Handled elsewhere
  if(FXScrollArea::onDNDMotion(sender,sel,ptr)) return 1;

  // Correct drop type
  if(offeredDNDType(FROM_DRAGNDROP,textType)){

    // Is target editable?
		if (!pdoc->IsReadOnly()) {
      action=inquireDNDAction();

      // Check for legal DND action
      if(action==DRAG_COPY || action==DRAG_MOVE){
				Point npt(ev->win_x, ev->win_y);
				pos = PositionFromLocation(npt);
				if (!inDragDrop) {
					inDragDrop = true;
					ptMouseLastBeforeDND = ptMouseLast;
				}
				ptMouseLast = npt;
				SetDragPosition(pos);
				if (PositionInSelection(pos) != 0) {
          acceptDrop(DRAG_ACCEPT);
        }
      }
    }
    return 1;
  }

  // Didn't handle it here
  return 0;
}


// Handle drag-and-drop drop
long ScintillaFOX::onDNDDrop(FXObject* sender,FXSelector sel,void* ptr){
  FXuchar *data,*junk; FXuint len,dum;

  // Stop scrolling
  stopAutoScroll();

  // Try handling it in base class first
  if(FXScrollArea::onDNDDrop(sender,sel,ptr)) return 1;

  // Should really not have gotten this if non-editable
	if (!pdoc->IsReadOnly()) {

	  // Try handle here
	  if(getDNDData(FROM_DRAGNDROP,textType,data,len)){
			dragWasDropped = true;
	    FXRESIZE(&data,FXchar,len+1); data[len]='\0';

	    // Need to ask the source to delete his copy
  	  if(inquireDNDAction()==DRAG_MOVE){
    	  getDNDData(FROM_DRAGNDROP,deleteType,junk,dum);
      	FXASSERT(!junk);
	    }

	    // Insert the new text
			bool isRectangular = ((len > 0) &&
					(data[len] == 0 && data[len-1] == '\n'));
			DropAt(posDrop, (const char *)(data), false, isRectangular);
  	
    	FXFREE(&data);
	  }
		else if (getDNDData(FROM_DRAGNDROP,urilistType,data,len)) {
			dragWasDropped = true;
	    FXRESIZE(&data,FXchar,len+1); data[len]='\0';
			NotifyURIDropped((FXchar *)data);
		}
  	return 1;
	}
  return 0;
}

// Service requested DND data
long ScintillaFOX::onDNDRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data;

  // Perhaps the target wants to supply its own data
  if(FXScrollArea::onDNDRequest(sender,sel,ptr)) return 1;

  // Return dragged text
  if(event->target==textType){
		if (primary.s == NULL) {
			CopySelectionRange(&primary);

		}

    setDNDData(FROM_DRAGNDROP,stringType,(FXuchar *)strdup(primary.s),strlen(primary.s));
    return 1;
    }

  // Delete dragged text
  if(event->target==deleteType){
		if (!pdoc->IsReadOnly()) {
			if (isDragging()) {
				int selStart = SelectionStart();
				int selEnd = SelectionEnd();
				if (posDrop > selStart) {
					if (posDrop > selEnd)
						posDrop = posDrop - (selEnd-selStart);
					else
						posDrop = selStart;
					posDrop = pdoc->ClampPositionIntoDocument(posDrop);
				}
			}
      ClearSelection();
      }
    return 1;
    }

  return 0;
  }

// ********************************************************************
// Selection
// ********************************************************************

// We lost the selection somehow
long ScintillaFOX::onSelectionLost(FXObject* sender,FXSelector sel,void* ptr){
  FXbool hadselection=hasSelection();
  FXScrollArea::onSelectionLost(sender,sel,ptr);
  if (hadselection)
		UnclaimSelection();
  return 1;
  }


// Somebody wants our selection
long ScintillaFOX::onSelectionRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;

  // Perhaps the target wants to supply its own data for the selection
  if(FXScrollArea::onSelectionRequest(sender,sel,ptr)) return 1;

  // Return text of the selection
  if(event->target==stringType){
		if (primary.s == NULL) {
			CopySelectionRange(&primary);
		}

    setDNDData(FROM_SELECTION,stringType,(FXuchar *)strdup(primary.s),strlen(primary.s));
    return 1;
  }

  return 0;
}

void ScintillaFOX::ClaimSelection()
{
  // Acquire selection
	if (currentPos != anchor) {
    acquireSelection(&stringType,1);
		primarySelection = true;
		primary.Set(0, 0);
	}
}

void ScintillaFOX::UnclaimSelection()
{
	//Platform::DebugPrintf("UnclaimPrimarySelection\n");
	if (!hasSelection()) {
		primary.Set(0, 0);
		primarySelection = false;
		FullPaint();
	}
}

void ScintillaFOX::ReceivedSelection(FXDNDOrigin origin)
{
  FXuchar *data; FXuint len;
	if (!pdoc->IsReadOnly()) {
    if(getDNDData(origin, stringType, data, len)) {
      FXRESIZE(&data,FXuchar,len+1); data[len]='\0';
			pdoc->BeginUndoAction();
			// Check for "\n\0" ending to string indicating that selection is rectangular
			int selStart = SelectionStart();
	    if(hasSelection() && (origin == FROM_CLIPBOARD)) {
				ClearSelection();
			}
// <FIXME/>
/*			bool isRectangular = ((len > 0) &&
				(data[len] == 0 && data[len - 1] == '\n'));
			if (isRectangular) {
				cout << data << flush;
				PasteRectangular(selStart, (const char *)data, len);
			} else */{
				pdoc->InsertString(currentPos, (const char *)data, len);
				SetEmptySelection(currentPos + len);
			}
			pdoc->EndUndoAction();
      FXFREE(&data);
		}
	}
}

// ********************************************************************
// Clipboard
// ********************************************************************

// We lost the clipboard somehow
long ScintillaFOX::onClipboardLost(FXObject* sender,FXSelector sel,void* ptr){
  FXScrollArea::onClipboardLost(sender,sel,ptr);
	//Platform::DebugPrintf("Clipboard lost\n");
	copyText.Set(0, 0);

  return 1;
}


// Somebody wants our clipboard
long ScintillaFOX::onClipboardRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data;

  // Try handling it in base class first
  if(FXScrollArea::onClipboardRequest(sender,sel,ptr)) return 1;

  // Requested data from clipboard
  if(event->target==stringType){
    setDNDData(FROM_CLIPBOARD,stringType,(FXuchar *)strdup(copyText.s),strlen(copyText.s));
    return 1;
	}

  return 0;
}

// ********************************************************************
// Keyboard
// ********************************************************************

// Map the keypad keys to their equivalent functions
static int KeyTranslate(int keyIn) {
	switch (keyIn) {
		case KEY_ISO_Left_Tab:
			return SCK_TAB;
		case KEY_KP_Down:
			return SCK_DOWN;
		case KEY_KP_Up:
			return SCK_UP;
		case KEY_KP_Left:
			return SCK_LEFT;
		case KEY_KP_Right:
			return SCK_RIGHT;
		case KEY_KP_Home:
			return SCK_HOME;
		case KEY_KP_End:
			return SCK_END;
		case KEY_KP_Page_Up:
			return SCK_PRIOR;
		case KEY_KP_Page_Down:
			return SCK_NEXT;
		case KEY_KP_Delete:
			return SCK_DELETE;
		case KEY_KP_Insert:
			return SCK_INSERT;
		case KEY_KP_Enter:
			return SCK_RETURN;
			
		case KEY_Down:
			return SCK_DOWN;
		case KEY_Up:
			return SCK_UP;
		case KEY_Left:
			return SCK_LEFT;
		case KEY_Right:
			return SCK_RIGHT;
		case KEY_Home:
			return SCK_HOME;
		case KEY_End:
			return SCK_END;
		case KEY_Page_Up:
			return SCK_PRIOR;
		case KEY_Page_Down:
			return SCK_NEXT;
		case KEY_Delete:
			return SCK_DELETE;
		case KEY_Insert:
			return SCK_INSERT;
		case KEY_Escape:
			return SCK_ESCAPE;
		case KEY_BackSpace:
			return SCK_BACK;
		case KEY_Tab:
			return SCK_TAB;
		case KEY_Return:
			return SCK_RETURN;
		case KEY_KP_Add:
			return SCK_ADD;
		case KEY_KP_Subtract:
			return SCK_SUBTRACT;
		case KEY_KP_Divide:
			return SCK_DIVIDE;
		default:
			return keyIn;
	}
}

long ScintillaFOX::onKeyPress(FXObject* sender,FXSelector sel,void* ptr)
{
  // Try handling it in base class first
  if (FXScrollArea::onKeyPress(sender,sel,ptr)) return 1;

  FXEvent *event=(FXEvent*)ptr;
	bool shift = event->state & SHIFTMASK;
	bool ctrl = event->state & CONTROLMASK;
	bool alt = event->state & ALTMASK;
	int key = event->code;
	if (ctrl && (key < 128))
		key = toupper(key);
	else if (!ctrl && (key >= KEY_KP_Multiply && key <= KEY_KP_9))
		key &= 0x7F;
	// Hack for keys over 256 and below command keys but makes Hungarian work.
	// This will have to change for Unicode
	else if ((key >= 0x100) && (key < 0x1000))
		key &= 0xff;
	else	
		key = KeyTranslate(key);

	bool consumed = false;
	int added = KeyDown(key, shift, ctrl, alt, &consumed);
	if (!consumed)
		consumed = added;
	//Platform::DebugPrintf("SK-key: %d %x %x\n",event->code, event->state, consumed);
	return consumed ? 1 : 0;
}

void ScintillaFOX::NotifyKey(int key, int modifiers) {
	SCNotification scn;
	scn.nmhdr.code = SCN_KEY;
	scn.ch = key;
	scn.modifiers = modifiers;

	NotifyParent(scn);
}

void ScintillaFOX::NotifyURIDropped(const char *list) {
	SCNotification scn;
	scn.nmhdr.code = SCN_URIDROPPED;
	scn.text = list;

	NotifyParent(scn);
}

int ScintillaFOX::KeyDefault(int key, int modifiers) {
	if (!(modifiers & SCI_CTRL) && !(modifiers & SCI_ALT) && (key < 256)) {
		AddChar(key);
		return 1;
	} else {
		// Pass up to container in case it is an accelerator
		NotifyKey(key, modifiers);
		return 0;
	}
	//Platform::DebugPrintf("SK-key: %d %x %x\n",key, modifiers);
}

// ********************************************************************

void ScintillaFOX::Initialise()
{
	SetTicking(true);
}

void ScintillaFOX::Finalise()
{
	SetTicking(false);
	ScintillaBase::Finalise();
}

void ScintillaFOX::SetVerticalScrollPos()
{
	pos_y = -topLine * vs.lineHeight;
	update();
}

void ScintillaFOX::SetHorizontalScrollPos()
{
	pos_x = -xOffset * getContentWidth() / 2000;
	update();
}

void ScintillaFOX::Copy()
{
  if (hasSelection()) {
    if (acquireClipboard(&stringType, 1)) {
			CopySelectionRange(&copyText);
    }
	}
}

void ScintillaFOX::Paste()
{
	ReceivedSelection(FROM_CLIPBOARD);
}

void ScintillaFOX::NotifyChange()
{
	target->handle(this, MKUINT(message, SEL_CHANGED), NULL);
}

void ScintillaFOX::NotifyParent(SCNotification scn)
{
	scn.nmhdr.hwndFrom = wMain.GetID();
	scn.nmhdr.idFrom = ctrlID;
	if (target)
		target->handle(this, MKUINT(message, SEL_COMMAND), &scn);
}

void ScintillaFOX::SetTicking(bool on)
{
	if (timer.ticking != on) {
		timer.ticking = on;
		if (timer.ticking) {
			timer.tickerID = FXApp::instance()->addTimeout(timer.tickSize, this, ID_TICK);
		} else {
			FXApp::instance()->removeTimeout(static_cast<FXTimer *>(timer.tickerID));
		}
	}
	timer.ticksToWait = caret.period;
}

void ScintillaFOX::SetMouseCapture(bool on)
{
	if (mouseDownCaptures) {
		if (on) {
			grab();
		} else {
			ungrab();
		}
	}
}

bool ScintillaFOX::HaveMouseCapture()
{
	return grabbed();
}

void ScintillaFOX::StartDrag()
{
	tryDrag = true;
	dragWasDropped = false;
}

sptr_t ScintillaFOX::DefWndProc(unsigned int, uptr_t, sptr_t)
{
	return 0;
}

void ScintillaFOX::CreateCallTipWindow(PRectangle rc)
{
}

void ScintillaFOX::AddToPopUp(const char * label, int cmd = 0, bool enabled = true)
{
	if (label[0]) {
		FXMenuCommand * item = new SCPopupMenuCommand(PMenu(popup.GetID()), label, NULL, this, ID(cmd));
		if (!enabled)
			item->disable();
	}
	else
		new FXMenuSeparator(PMenu(popup.GetID()));
}

sptr_t ScintillaFOX::DirectFunction(
    ScintillaFOX *sciThis, unsigned int iMessage, uptr_t wParam, sptr_t lParam) {
	return sciThis->WndProc(iMessage, wParam, lParam);
}

FXint ScintillaFOX::getContentWidth()
{
	return getWidth() * 10;
}

FXint ScintillaFOX::getContentHeight()
{
	return (cs.LinesDisplayed() + 1) * vs.lineHeight;
}

PRectangle ScintillaFOX::GetClientRectangle() {
	PRectangle rc(0, 0, getViewportWidth() - 1, getViewportHeight() - 1);
	if (horizontal->shown())
		rc.bottom -= horizontal->getDefaultHeight();
	if (vertical->shown())
		rc.right -= vertical->getDefaultWidth();
	return rc;
}

void ScintillaFOX::ReconfigureScrollBars()
{
	if (horizontalScrollBarVisible)
		options &= ~HSCROLLER_NEVER;
	else
		options |= HSCROLLER_NEVER;
	recalc();
}

bool ScintillaFOX::ModifyScrollBars(int, int)
{
	bool modified = false;
	int line = vs.lineHeight;
	if (vertical->getLine() != line) {
		vertical->setLine(line);
		modified = true;
	}
	if (horizontal->getLine() != 20) {
		horizontal->setLine(20);
		modified = true;
	}
	return modified;
}

void ScintillaFOX::moveContents(FXint x,FXint y)
{
	int line = (-y + vs.lineHeight / 2) / vs.lineHeight;
	if (line != topLine) {
		pos_y = -line * vs.lineHeight;
		ScrollTo(line);
		update();
	}
	if (x != pos_x) {
		pos_x = x;
		HorizontalScrollTo(-x * 2000 / getContentWidth());
		update();
	}
}

long ScintillaFOX::onConfigure(FXObject *sender, FXSelector sel, void * ptr)
{
	DropGraphics();
	return FXScrollArea::onConfigure(sender, sel, ptr);
}

// --------------------------------------------------------------------
// Synchronously paint a rectangle of the window.
void ScintillaFOX::SyncPaint(PRectangle rc) {
	paintState = painting;
	rcPaint = rc;
	PRectangle rcText = GetTextRectangle();
	paintingAllText = rcPaint.Contains(rcText);
	//Platform::DebugPrintf("ScintillaGTK::SyncPaint %0d,%0d %0d,%0d\n",
	//	rcPaint.left, rcPaint.top, rcPaint.right, rcPaint.bottom);
	Surface *sw = Surface::Allocate();
	if (sw) {
		sw->Init(PWidget(wMain));
		Paint(sw, rcPaint);
		sw->Release();
		delete sw;
	}
	if (paintState == paintAbandoned) {
		// Painting area was insufficient to cover new styling or brace highlight positions
		FullPaint();
	}
	paintState = notPainting;
}

// Redraw all of text area. This paint will not be abandoned.
void ScintillaFOX::FullPaint() {
	paintState = painting;
	rcPaint = GetTextRectangle();
	//Platform::DebugPrintf("ScintillaGTK::FullPaint %0d,%0d %0d,%0d\n",
	//	rcPaint.left, rcPaint.top, rcPaint.right, rcPaint.bottom);
	paintingAllText = true;
	Surface *sw = Surface::Allocate();
	if (sw) {
		sw->Init(PWidget(wMain));
		Paint(sw, rcPaint);
		sw->Release();
		delete sw;
	}
	paintState = notPainting;
}

// ====================================================================
ScintillaObject::ScintillaObject(FXComposite * p, FXObject * target, FXSelector sel)
{
	_scint = new ScintillaFOX(p, target, sel);
}

ScintillaObject::~ScintillaObject()
{
	delete _scint;
}

FXWindow * ScintillaObject::FX()
{
	return _scint;
}

ScintillaBase * ScintillaObject::SB()
{
	return _scint;
}

void ScintillaObject::SetID(int id)
{
	_scint->ctrlID = id;
}

sptr_t scintilla_send_message(ScintillaFOX *sci, unsigned int iMessage, uptr_t wParam, sptr_t lParam) {
	return sci->WndProc(iMessage, wParam, lParam);
}

